import base64
from contextlib import contextmanager
from typing import Dict, Iterator

from laituri.docker.credential_manager.docker_v1 import docker_v1_credential_manager
from laituri.docker.credential_manager.errors import ECRLoginFailed
from laituri.types import LogStatusCallable, RegistryCredentialsDict


@contextmanager
def ecr_with_role_v1_credential_manager(
    *,
    image: str,
    registry_credentials: RegistryCredentialsDict,
    log_status: LogStatusCallable,
) -> Iterator[None]:
    from boto3 import Session
    role_name = registry_credentials['role_name']
    creds = get_role_credentials_from_instance_metadata(role_name)
    try:
        sess = Session(
            region_name=registry_credentials['region'],
            aws_access_key_id=creds['access_key'],
            aws_secret_access_key=creds['secret_key'],
            aws_session_token=creds['token'],
        )
        ecr = sess.client('ecr')
        result = ecr.get_authorization_token()
        auth_token = base64.b64decode(result['authorizationData'][0]['authorizationToken']).decode()
        username, password = auth_token.split(':')
        docker_credentials = {
            'username': username,
            'password': password,
        }
    except Exception as exc:
        raise ECRLoginFailed(f"Role based docker credentials fetch failed: {exc}") from exc

    with docker_v1_credential_manager(image=image, registry_credentials=docker_credentials, log_status=log_status):
        yield


def get_role_credentials_from_instance_metadata(role_name: str) -> Dict[str, str]:
    from botocore.utils import InstanceMetadataFetcher
    fetcher = InstanceMetadataFetcher()
    creds: Dict[str, str] = fetcher.retrieve_iam_role_credentials()
    if creds.get('role_name') != role_name:
        # User tried to use a IAM role which is not attached to the EC2 machine
        raise ECRLoginFailed(f"No instance role {role_name} found")
    return creds



def get_image_hostname(image: str) -> str:
    if image.count('/'):
        first = image.split('/')[0]
        if '.' in first or ':' in first or first == 'localhost':
            return first
    return 'docker.io'



from typing import Optional

from requests import RequestException, Response

from laituri.errors import LaituriError


class DockerLoginFailed(LaituriError):
    """We failed to login to Docker from some reason."""

    pass


class ECRLoginFailed(LaituriError):
    """We failed to login to ECR from some reason."""

    pass


class InvalidDockerCommand(LaituriError):
    """The Docker command is misconfigured."""


class CallbackFailed(LaituriError):
    """
    Calling a registry credentials callback failed.
    """

    def get_callback_response(self) -> Optional[Response]:
        """
        Attempt to reach into the inner exception to get the response object returned by the callback.
        """
        if isinstance(self.__cause__, RequestException):
            response = self.__cause__.response
            if isinstance(response, Response):
                return response
        return None



import logging
import shutil
import subprocess
from contextlib import contextmanager
from typing import Iterator

from laituri import settings
from laituri.docker.credential_manager.errors import DockerLoginFailed, InvalidDockerCommand
from laituri.types import LogStatusCallable, RegistryCredentialsDict
from laituri.utils.images import get_image_hostname

log = logging.getLogger(__name__)


@contextmanager
def docker_v1_credential_manager(
    *,
    image: str,
    registry_credentials: RegistryCredentialsDict,
    log_status: LogStatusCallable,
) -> Iterator[None]:
    # If image looks like a dockerhub image, use docker.io as the registry domain
    hostname = get_image_hostname(image)
    try:
        docker_login(
            domain=str(hostname),
            username=str(registry_credentials['username']),
            password=str(registry_credentials['password']),
        )
    except DockerLoginFailed as dlf:
        raise DockerLoginFailed(f'Failed Docker login to {hostname}: {str(dlf)}') from dlf
    try:
        yield
    finally:
        docker_logout(hostname)


def get_docker_command() -> str:
    cmd = shutil.which(settings.DOCKER_COMMAND)
    if not cmd:
        raise InvalidDockerCommand(f"Invalid Docker command: {cmd}")
    return cmd


def docker_login(domain: str, username: str, password: str) -> bool:
    """
    Use Docker command-line client to login to the specified image registry.
    """
    args = [
        get_docker_command(),
        'login',
        '--username', username,
        '--password-stdin',
        domain,
    ]
    log.debug(f"Running `{' '.join(args)}`")
    proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
        cmd_input = (password + '\n').encode('utf-8')
        stdout, _ = proc.communicate(input=cmd_input, timeout=settings.DOCKER_TIMEOUT)
    except subprocess.TimeoutExpired as te:
        raise DockerLoginFailed('timed out') from te
    if proc.returncode != 0:
        raise DockerLoginFailed(stdout.decode('utf-8', errors='ignore'))
    return True


def docker_logout(domain: str) -> None:
    """
    Use Docker command-line client to logout from the specified image registry.
    """
    # when logging out of "docker.io" they actually mean "https://index.docker.io/v1/"
    if domain == 'docker.io':
        domain = 'https://index.docker.io/v1/'

    try:
        log.debug(f'Running `docker logout {domain}`')
        subprocess.check_call([
            get_docker_command(),
            'logout',
            domain,
        ], stderr=subprocess.STDOUT, stdout=subprocess.PIPE, timeout=settings.DOCKER_TIMEOUT)
    except subprocess.CalledProcessError as cpe:
        message = cpe.stdout.decode('utf-8', errors='ignore')
        log.warning(f'Failed `docker logout {domain}`: {message}')



from contextlib import contextmanager
from typing import ContextManager, Iterator

from laituri.types import LogStatusCallable, RegistryCredentialsDict


@contextmanager
def dummy_credential_manager(
    *,
    image: str,
    registry_credentials: RegistryCredentialsDict,
    log_status: LogStatusCallable,
) -> Iterator[None]:
    """
    Credential context manager that does nothing.

    This can used when no credentials are included.
    """
    yield


def get_dummy_credential_manager() -> ContextManager[None]:
    """
    Construct a dummy credential manager without having to think about arguments.
    """
    return dummy_credential_manager(image="", registry_credentials={}, log_status=lambda s: None)



from typing import Dict, Tuple

from laituri.docker.credential_manager.docker_v1 import docker_v1_credential_manager
from laituri.docker.credential_manager.ecr_with_role_v1 import ecr_with_role_v1_credential_manager
from laituri.docker.credential_manager.registry_credentials_callback_v1 import (
    registry_credentials_callback_v1_credential_manager,
)
from laituri.types import CredentialManagerFactory

NameAndVersion = Tuple[str, str]

credential_managers: Dict[NameAndVersion, CredentialManagerFactory] = {
    ('docker', '1'): docker_v1_credential_manager,
    ('registry-credentials-callback', '1'): registry_credentials_callback_v1_credential_manager,
    ('ecr-with-role', '1'): ecr_with_role_v1_credential_manager,
}



from typing import Any, Callable, ContextManager, Dict

LogStatusCallable = Callable[[str], None]
RegistryCredentialsDict = Dict[str, Any]
CredentialManager = ContextManager[None]
CredentialManagerFactory = Callable[..., CredentialManager]



from typing import Optional

from laituri.docker.credential_manager.credential_managers import credential_managers
from laituri.docker.credential_manager.dummy import get_dummy_credential_manager
from laituri.types import CredentialManager, LogStatusCallable, RegistryCredentialsDict


def _noop_log_status(message: str) -> None:
    pass  # pragma: nocover


def get_credential_manager(
    *,
    image: str,
    registry_credentials: Optional[RegistryCredentialsDict] = None,
    log_status: LogStatusCallable = _noop_log_status,
) -> CredentialManager:
    """
    Get a credential context manager object based on the registry credentials dictionary passed in.

    If registry credentials are missing or incomplete, will still trigger the wrapped action.
    If registry credentials are invalid, raises DockerLoginFailed.

    :param image: full Docker image name, including the registry domain and tag if applicable
    :param registry_credentials: optional {type, version, username, password} dict for registry login
    :param log_status: optional function to use for user-facing status logging
    :raises DockerLoginFailed
    :return: ContextManager
    """
    if not isinstance(registry_credentials, dict):  # no credentials to manage
        return get_dummy_credential_manager()

    credentials_type = registry_credentials.get('type')
    version = registry_credentials.get('version')
    manager = credential_managers.get((str(credentials_type), str(version)))
    if manager:
        return manager(
            image=image,
            registry_credentials=registry_credentials,
            log_status=log_status,
        )

    log_status(f'Unable to parse {type} version {version} registry credentials; upcoming action may fail.')
    return get_dummy_credential_manager()



import random
import time
from functools import wraps
from typing import Any, Callable, TypeVar, cast

T = TypeVar('T', bound=Callable[..., Any])


def retry(*, tries: int = 5, max_delay: float = 32) -> Callable[[T], T]:
    """
    Decorator that retries the wrapped function if any exception occurs.
    """

    def inner_retry(func: T) -> T:
        @wraps(func)
        def wrapped_func(*args, **kwargs):  # type: ignore
            attempt = 1
            while attempt < tries:
                try:
                    return func(*args, **kwargs)
                except Exception:
                    delay = (2 ** (attempt - 1))  # 1, 2, 4, 8, 16, 32...
                    delay += random.random()  # a tiny bit of random for desynchronizing multiple potential users
                    delay = min(delay, max_delay)
                    time.sleep(delay)
                    attempt += 1
            # and finally, just try one more time but let any exceptions bubble up
            return func(*args, **kwargs)

        return cast(T, wrapped_func)

    return inner_retry



from contextlib import contextmanager
from typing import Any, Dict, Iterator

import requests
from requests.utils import default_headers

import laituri
from laituri.docker.credential_manager.docker_v1 import docker_v1_credential_manager
from laituri.docker.credential_manager.errors import CallbackFailed
from laituri.types import LogStatusCallable, RegistryCredentialsDict
from laituri.utils.retry import retry


@contextmanager
def registry_credentials_callback_v1_credential_manager(
    *,
    image: str,
    registry_credentials: RegistryCredentialsDict,
    log_status: LogStatusCallable,
) -> Iterator[None]:
    try:
        docker_credentials = fetch_docker_credentials(registry_credentials)
    except Exception as exc:
        raise CallbackFailed(f"Credential callback failed: {exc}") from exc

    with docker_v1_credential_manager(image=image, registry_credentials=docker_credentials, log_status=log_status):
        yield


@retry()
def fetch_docker_credentials(request_info: Dict[str, Any]) -> RegistryCredentialsDict:
    headers = default_headers()  # type: ignore[no-untyped-call]
    headers['User-Agent'] = f'{headers.get("User-Agent")} laituri/{laituri.__version__}'
    ri_headers = request_info.get('headers')
    if isinstance(ri_headers, dict):
        headers.update({key: str(value) for (key, value) in ri_headers.items() if key and value})

    response = requests.request(
        method=request_info.get('method', 'POST'),
        url=request_info['url'],
        headers=headers,
        timeout=15,
    )
    response.raise_for_status()
    data = response.json()
    if not isinstance(data, dict):
        raise ValueError(f"Invalid response data: {data}")
    return {str(k): v for (k, v) in data.items()}
